/*
  ==============================================================================

    This file was auto-generated!

    It contains the basic framework code for a JUCE plugin processor.

  ==============================================================================
*/

#include "PluginProcessor.h"
#include "PluginEditor.h"
#include "math.h"

//==============================================================================
slovikp04AudioProcessor::slovikp04AudioProcessor()
#ifndef JucePlugin_PreferredChannelConfigurations
     : AudioProcessor (BusesProperties()
                     #if ! JucePlugin_IsMidiEffect
                      #if ! JucePlugin_IsSynth
                       .withInput  ("Input",  AudioChannelSet::stereo(), true)
                      #endif
                       .withOutput ("Output", AudioChannelSet::stereo(), true)
                     #endif
                       )
#endif
{
    addParameter (delayTimeLUParam = new AudioParameterFloat ("delayTimeL", // parameterID
                                                              "delayTimeL", // parameter name
                                                              NormalisableRange<float> (0.0, 2000.0), //parameter range
                                                              0.0f)); // default value
    
    addParameter (delayTimeRUParam = new AudioParameterFloat ("delayTimeR", // parameterID
                                                              "delayTimeR", // parameter name
                                                              NormalisableRange<float> (0.0, 2000.0), //parameter range
                                                              0.0f)); // default value
    
    addParameter (dryWetUParam = new AudioParameterFloat ("dryWet", // parameterID
                                                         "dryWet", // parameter name
                                                         NormalisableRange<float> (0.0, 100.0f), //parameter range
                                                         0.0f)); // default value
    
    addParameter (oGainUParam = new AudioParameterFloat ("outGain", // parameterID
                                                             "outGain", // parameter name
                                                             NormalisableRange<float> (-6.0, 6.0f), //parameter range
                                                             0.0f)); // default value
    
    addParameter (iGainUParam = new AudioParameterFloat ("inGain", // parameterID
                                                         "inGain", // parameter name
                                                         NormalisableRange<float> (-6.0, 6.0f), //parameter range
                                                         0.0f)); // default value
    
    addParameter (fbUParam = new AudioParameterFloat ("feedback", // parameterID
                                                         "feedback", // parameter name
                                                         NormalisableRange<float> (0.0, 99.0f), //parameter range
                                                         0.0f)); // default value
    
    addParameter (link = new AudioParameterFloat ("link", // parameterID
                                                  "Link", // parameter name
                                                  NormalisableRange<float> (0.0f, 1.0f), //parameter range
                                                  1.0f)); // default value
    
}

slovikp04AudioProcessor::~slovikp04AudioProcessor()
{
}

//==============================================================================
const String slovikp04AudioProcessor::getName() const
{
    return JucePlugin_Name;
}

bool slovikp04AudioProcessor::acceptsMidi() const
{
   #if JucePlugin_WantsMidiInput
    return true;
   #else
    return false;
   #endif
}

bool slovikp04AudioProcessor::producesMidi() const
{
   #if JucePlugin_ProducesMidiOutput
    return true;
   #else
    return false;
   #endif
}

bool slovikp04AudioProcessor::isMidiEffect() const
{
   #if JucePlugin_IsMidiEffect
    return true;
   #else
    return false;
   #endif
}

double slovikp04AudioProcessor::getTailLengthSeconds() const
{
    return 0.0;
}

int slovikp04AudioProcessor::getNumPrograms()
{
    return 1;   // NB: some hosts don't cope very well if you tell them there are 0 programs,
                // so this should be at least 1, even if you're not really implementing programs.
}

int slovikp04AudioProcessor::getCurrentProgram()
{
    return 0;
}

void slovikp04AudioProcessor::setCurrentProgram (int index)
{
}

const String slovikp04AudioProcessor::getProgramName (int index)
{
    return {};
}

void slovikp04AudioProcessor::changeProgramName (int index, const String& newName)
{
}

//==============================================================================
void slovikp04AudioProcessor::prepareToPlay (double sampleRate, int samplesPerBlock)
{
    // Use this method as the place to do any pre-playback
    // initialisation that you need..
    
    fs = sampleRate; // samples per second
    leftDelayMsec = *delayTimeLUParam;
    rightDelayMsec = *delayTimeRUParam;
    
    
    // initialize delay check
    curDelayL = delaySampsL;
    curDelayR = delaySampsR;
    prevDelayL = curDelayL;
    prevDelayR = curDelayR;
    
    // initialize maximum delay based on max delay time desired
    maxDelayInSamps = fs * 2; // 2 seconds (or 2000 ms) of maximum delay
    delayL.setMaximumDelay(maxDelayInSamps); // initializing delay
    delayR.setMaximumDelay(maxDelayInSamps); // initializing delay
    
    delayL.setDelay(maxDelayInSamps);
    delayR.setDelay(maxDelayInSamps);
    
}

void slovikp04AudioProcessor::releaseResources()
{
    // When playback stops, you can use this as an opportunity to free up any
    // spare memory, etc.
    
    // clear the delay line here to avoid leftover delay that creates pops
    // when starting or stopping playback
    for (int i = 0; i < maxDelayInSamps; i++) {
        delayL.tick(0.0); // tick through each sample and set delay to zero
        delayR.tick(0.0);
    }
    
}

#ifndef JucePlugin_PreferredChannelConfigurations
bool slovikp04AudioProcessor::isBusesLayoutSupported (const BusesLayout& layouts) const
{
  #if JucePlugin_IsMidiEffect
    ignoreUnused (layouts);
    return true;
  #else
    // This is the place where you check if the layout is supported.
    // In this template code we only support mono or stereo.
    if (layouts.getMainOutputChannelSet() != AudioChannelSet::mono()
     && layouts.getMainOutputChannelSet() != AudioChannelSet::stereo())
        return false;

    // This checks if the input layout matches the output layout
   #if ! JucePlugin_IsSynth
    if (layouts.getMainOutputChannelSet() != layouts.getMainInputChannelSet())
        return false;
   #endif

    return true;
  #endif
}
#endif

void slovikp04AudioProcessor::calcAlgorithmParams()
{
    // grab UI to alg variable
    leftDelayMsec = *delayTimeLUParam;
    rightDelayMsec = *delayTimeRUParam;
    
    if (*link == 0) {
        leftDelayMsec = rightDelayMsec; // link parameters when button is pressed
    } else {
        leftDelayMsec = *delayTimeLUParam;
        rightDelayMsec = *delayTimeRUParam;
    }

    // convert delay from ms to samples
    delaySampsL = calcDelaySampsFromMsec(leftDelayMsec); // converted to samples
    delaySampsR = calcDelaySampsFromMsec(rightDelayMsec); // converted to samples
    // set current delay
    curDelayL = delaySampsL;
    curDelayR = delaySampsR;
    
    fbAParam = *fbUParam * 0.01; // convert user % to 0 - 0.99
    
    dryWetAParam = *dryWetUParam * 0.01; // convert user % to 0 - 1
    dGain = 1 - dryWetAParam; // if mix knob is 100%, dry = 0
    wGain = dryWetAParam; // if mix knob is 100%, wet = 1
    
    // using dB
    oGainAParam = pow(10.0,(*oGainUParam)/20.0); //convert user out dB gain to linear gain
    iGainAParam = pow(10.0,(*iGainUParam)/20.0); //convert user in dB gain to linear gain

}

//linear interpolation
float slovikp04AudioProcessor::linInterpolation(float start, float end, float fract) {
    return start + ( fract * (end - start)); //a - fract * ( b - a )
}

float slovikp04AudioProcessor::calcDelaySampsFromMsec (float msec) {
    return fs * (msec/1000);
}

void slovikp04AudioProcessor::processBlock (AudioBuffer<float>& buffer, MidiBuffer& midiMessages)
{
    ScopedNoDenormals noDenormals;
    auto totalNumInputChannels  = getTotalNumInputChannels();
    auto totalNumOutputChannels = getTotalNumOutputChannels();

    // In case we have more outputs than inputs, this code clears any output
    // channels that didn't contain input data, (because these aren't
    // guaranteed to be empty - they may contain garbage).
    // This is here to avoid people getting screaming feedback
    // when they first compile a plugin, but obviously you don't need to keep
    // this code if your algorithm always overwrites all the output channels.
    for (auto i = totalNumInputChannels; i < totalNumOutputChannels; ++i)
        buffer.clear (i, 0, buffer.getNumSamples());
    
    calcAlgorithmParams();
    const int numSamps = buffer.getNumSamples();

    float delOutL, delOutR;
    float tmpL, tmpR;
    float* leftChannel;
    float* rightChannel;
    float fract;
    // This is the place where you'd normally do the guts of your plugin's
    // audio processing...
    // Make sure to reset the state if your inner loop is processing
    // the samples and the outer loop is handling the channels.
    // Alternatively, you can process the samples with the channels
    // interleaved by keeping the same state.
    for (int channel = 0; channel < totalNumInputChannels; ++channel)
    {
        auto* channelData = buffer.getWritePointer (channel);
        if (channel == 0) { // left side
            leftChannel = channelData;
        } else if (channel == 1) { // right side
            rightChannel = channelData;
        }
        
        if (totalNumInputChannels < 2) { // check for mono
            rightChannel = leftChannel;
        }
        
        // ..do something to the data...
            for (int i = 0; i < numSamps; ++i)
            {
                fract = ((float) i / (float) numSamps);
                
                if (curDelayL != prevDelayL) {
                    leftDelayMsec = linInterpolation(prevDelayL, curDelayL, fract);
                    delaySampsL = calcDelaySampsFromMsec(leftDelayMsec); // converted to samples
                }
                
                if (curDelayR != prevDelayR) {
                    rightDelayMsec = linInterpolation(prevDelayR, curDelayR, fract);
                    delaySampsR = calcDelaySampsFromMsec(rightDelayMsec);
                }
                
                    // ping pong delay
                    delayL.setDelay(delaySampsL);
                    delayR.setDelay(delaySampsR);
                    
                    tmpL = delayL.nextOut() * fbAParam; // L feedback
                    tmpR = delayR.nextOut() * fbAParam; // R feedback
                
                    delOutL = delayL.tick((leftChannel[i] * iGainAParam) + tmpR); // dly in * inGain + R channel fb
                    leftChannel[i] = (leftChannel[i] * dGain) + (wGain * delOutL); // dry signal + wet signal
                    leftChannel[i] *= oGainAParam; // allow dB inc/dec after entire line
                
                    delOutR = delayR.tick((rightChannel[i] * iGainAParam) + tmpL); // dly in * inGain + L channel fb
                    rightChannel[i] = (rightChannel[i] * dGain) + (wGain * delOutR); // dry signal + wet signal
                    rightChannel[i] *= oGainAParam; // allow dB inc/dec after entire line
                }
        }
        prevDelayL = curDelayL;
        prevDelayR = curDelayR;
}

//==============================================================================
bool slovikp04AudioProcessor::hasEditor() const
{
    return true; // (change this to false if you choose to not supply an editor)
}

AudioProcessorEditor* slovikp04AudioProcessor::createEditor()
{
    return new slovikp04AudioProcessorEditor (*this);
}

//==============================================================================
void slovikp04AudioProcessor::getStateInformation (MemoryBlock& destData)
{
    // You should use this method to store your parameters in the memory block.
    // You could do that either as raw data, or use the XML or ValueTree classes
    // as intermediaries to make it easy to save and load complex data.
    
    //for multiple vars.
    std::unique_ptr<XmlElement> xml (new XmlElement ("DelayLParams"));
    xml->setAttribute ("delayTimeL", (double) *delayTimeLUParam);
    copyXmlToBinary (*xml, destData);
    
    (new XmlElement ("DelayRParams"));
    xml->setAttribute ("delayTimeR", (double) *delayTimeRUParam);
    copyXmlToBinary (*xml, destData);
    
    (new XmlElement ("outGain"));
    xml->setAttribute ("outGain", (double) *oGainUParam);
    copyXmlToBinary (*xml, destData);
    
    (new XmlElement ("inGain"));
    xml->setAttribute ("inGain", (double) *iGainUParam);
    copyXmlToBinary (*xml, destData);
    
    (new XmlElement ("dryWet"));
    xml->setAttribute ("dryWet", (double) *dryWetUParam);
    copyXmlToBinary (*xml, destData);
    
    (new XmlElement ("feedback"));
    xml->setAttribute ("feedback", (double) *fbUParam);
    copyXmlToBinary (*xml, destData);
}

void slovikp04AudioProcessor::setStateInformation (const void* data, int sizeInBytes)
{
    // You should use this method to restore your parameters from this memory block,
    // whose contents will have been created by the getStateInformation() call.
    
    //for multiple vars
    std::unique_ptr<XmlElement> xmlState (getXmlFromBinary (data, sizeInBytes));
    if (xmlState.get() != nullptr)
        if (xmlState->hasTagName ("DelayLParams"))
            *delayTimeLUParam = xmlState->getDoubleAttribute ("delayTimeL", 0.0);
    
    if (xmlState.get() != nullptr)
        if (xmlState->hasTagName ("DelayRParams"))
            *delayTimeRUParam = xmlState->getDoubleAttribute ("delayTimeR", 0.0);
    
    if (xmlState.get() != nullptr)
        if (xmlState->hasTagName ("outGain"))
            *oGainUParam = xmlState->getDoubleAttribute ("outGain", 0.0);
    
    if (xmlState.get() != nullptr)
        if (xmlState->hasTagName ("inGain"))
            *iGainUParam = xmlState->getDoubleAttribute ("inGain", 0.0);
    
    if (xmlState.get() != nullptr)
        if (xmlState->hasTagName ("dryWet"))
            *dryWetUParam = xmlState->getDoubleAttribute ("dryWet", 0.0);
    
    if (xmlState.get() != nullptr)
        if (xmlState->hasTagName ("feedback"))
            *fbUParam = xmlState->getDoubleAttribute ("feedback", 0.0);
}

//==============================================================================
// This creates new instances of the plugin..
AudioProcessor* JUCE_CALLTYPE createPluginFilter()
{
    return new slovikp04AudioProcessor();
}
